#include "conncomp.h"

void find_connected_components( 
							   IplImage*	mask, 
							   int			poly1_hull0, 
							   float			perimScale, 
							   int*			num, 
							   CvRect*		bbs, 
							   CvPoint*		centers 
							   ) 
{
	static CvMemStorage* 	mem_storage 	= NULL; 
	static CvSeq* 		contours 		= NULL;

	// Է ũ  ϰ Ѵ 
	// 
	cvMorphologyEx( mask, mask, 0, 0, CV_MOP_OPEN, CVCLOSE_ITR ); 
	cvMorphologyEx( mask, mask, 0, 0, CV_MOP_CLOSE, CVCLOSE_ITR ); 

	//ū  ѷ ܰ ã´ 
	// 
	if( mem_storage == NULL ) {
		mem_storage = cvCreateMemStorage( 0 ); 
	} else { 
		cvClearMemStorage( mem_storage ); 
	}

	CvContourScanner scanner = cvStartFindContours( 
		mask, 
		mem_storage, 
		sizeof(CvContour), 
		CV_RETR_EXTERNAL, 
		CV_CHAIN_APPROX_SIMPLE
	); 

	CvSeq* c; 
	int numCont = 0; 
	while( (c = cvFindNextContour( scanner )) != NULL ) 
	{
		double len = cvContourPerimeter( c );

		// ܰ (len) ּ Ӱ谪  
		// 
		double q = (mask->height + mask->width)/perimScale;

		// ܰ ̰ ʹ   Ѵ.
		// 
		if( len < q ) { 
			cvSubstituteContour( scanner, NULL ); 
		} else {
			// ܰ ̰  ū ,  Ѵ 
			// 
			CvSeq* c_new; 

			if( poly1_hull0 ) 
			{
				// ٰ ٻȭ 
				// 
				c_new = cvApproxPoly(
					c, 
					sizeof(CvContour), 
					mem_storage, 
					CV_POLY_APPROX_DP, 
					CVCONTOUR_APPROX_LEVEL, 
					0
					);
			} 
			else 
			{
				//   ٻȭ
				// 
				c_new = cvConvexHull2(
					c, 
					mem_storage, 
					CV_CLOCKWISE, 
					1
					); 
			} 
			cvSubstituteContour( scanner, c_new ); 
			numCont++;
		} 
	} 
	contours = cvEndFindContours( &scanner ); 

	// Ǹ   
	const CvScalar CVX_WHITE = CV_RGB(0xff, 0xff, 0xff);
	const CvScalar CVX_BLACK = CV_RGB(0x00, 0x00, 0x00);

	// ã   ׸. 
	// 
	cvZero( mask ); 
	IplImage *maskTemp;

	//  ߽ɰ ٿ ڽ Ѵ. 
	// 
	if( num != NULL ) 
	{
		//ڰ  ġ ϱ⸦ Ѵ. 
		// 
		int N = *num, numFilled = 0, i = 0; 
		CvMoments moments; 
		double M00, M01, M10; 
		maskTemp = cvCloneImage(mask); 

		for( i = 0, c = contours ; c != NULL ; c = c->h_next, i++ ) 
		{
			if( i < N )
			{ 
				// (*num) ۾ Ѵ. 
				// 
				cvDrawContours(
					maskTemp, 
					c, 
					CVX_WHITE, 
					CVX_WHITE, 
					-1, 
					CV_FILLED, 
					8
					);

				//  ܰ ߽ ã´. 
				// 
				if(centers != NULL) 
				{
					cvMoments(maskTemp,&moments,1); 
					M00 = cvGetSpatialMoment(&moments,0,0); 
					M10 = cvGetSpatialMoment(&moments,1,0); 
					M01 = cvGetSpatialMoment(&moments,0,1); 
					centers[i].x = (int)(M10/M00); 
					centers[i].y = (int)(M01/M00);
				}

				//  δ ٿ ڽ Ѵ. 
				// 
				if(bbs != NULL) {
					bbs[i] = cvBoundingRect(c); 
				} 

				cvZero(maskTemp); 
				numFilled++;
			} 

			// ä ܰ mask  ׸. 
			// 
			cvDrawContours(
				mask, 
				c, 
				CVX_WHITE, 
				CVX_WHITE, 
				-1, 
				CV_FILLED,
				8
				); 
		} 	// end of for 
		*num = numFilled; 
		cvReleaseImage( &maskTemp );
	} 

	// ƴϸ ܼ ũ ׸⸸ Ѵ. 
	// 
	else {
		// ڰ  ġ ϴ ,  ܰ ׸⸦ Ѵ. 
		// 
		for( c=contours; c != NULL; c = c->h_next ) {
			cvDrawContours( 
				mask, 
				c, 
				CVX_WHITE, 
				CVX_BLACK, 
				-1, 
				CV_FILLED, 
				8
				); 
		} 
	} 
}
